# Copyright (c) 2022, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals

import logging
from django.middleware.csrf import CsrfViewMiddleware
from django.utils.cache import patch_vary_headers
from rest_framework.authentication import get_authorization_header
from rest_framework.views import APIView

from .compat import six
from .perms import find_rule
from .utils import get_current_app


LOGGER = logging.getLogger(__name__)

ACCESS_CONTROL_ALLOW_HEADERS = 'Access-Control-Allow-Headers'
ACCESS_CONTROL_ALLOW_ORIGIN = 'Access-Control-Allow-Origin'
ACCESS_CONTROL_ALLOW_CREDENTIALS = 'Access-Control-Allow-Credentials'

ACCESS_CONTROL_ALLOW_HEADERS_ALLOWED = \
    "Origin, X-Requested-With, Content-Type, Accept, X-CSRFToken, Authorization"


class RulesMiddleware(CsrfViewMiddleware):
    """
    Disables CSRF check if the HTTP request is forwarded.
    Pass OPTIONS HTTP request regardless since authorization header is not
    sent along by browser (CORS).
    """

    @staticmethod
    def patch_set_cookies(response, domain):
        if not response.cookies:
            return
        for key in response.cookies:
            response.cookies[key]['domain'] = domain

    def process_view(self, request, callback, callback_args, callback_kwargs):
        view_class = getattr(callback, 'view_class', None)
        if hasattr(view_class, 'conditional_forward'):
            app = get_current_app(request)
            request.matched_rule, request.matched_params = find_rule(
                request, app)
            if (request.matched_rule and request.matched_rule.is_forward
                and request.method not in ('GET', 'HEAD', 'OPTIONS', 'TRACE')):
                # We are forwarding the request so the CSRF is delegated
                # to the application handling the forwarded request.
                #pylint:disable=protected-access
                request._dont_enforce_csrf_checks = True
                LOGGER.debug("dont enforce csrf checks on %s %s",
                    request.method, request.path)

        auth = get_authorization_header(request).split()
        if auth and auth[0].lower() in [b'basic', b'bearer']:
            # We need to support API calls from the command line.
            #pylint:disable=protected-access
            request._dont_enforce_csrf_checks = True
            LOGGER.debug("dont enforce csrf checks on %s %s because"\
                " we have an authorization header",
                request.method, request.path)

        # CORS will first send an OPTIONS request with no authorization header
        # and expect to get a 200 OK response.
        if request.method.lower() == 'options':
            if view_class and issubclass(view_class, APIView):
                # Duplicates what Django does before calling `dispatch`
                view = view_class(**callback.initkwargs)
                view.setup(request, *callback_args, **callback_kwargs)
                if not hasattr(view, 'request'):
                    raise AttributeError("%s instance has no 'request'"
                        " attribute. Did you override setup() and forget"
                        " to call super()?" % view_class.__name__)
                if hasattr(view, 'initialize_request'):
                    # Duplicates what rest_framework does before calling
                    # `options` minus testing for permissions since we won't
                    # have a user (no Authorization header).
                    request = view.initialize_request(
                        request, *callback_args, **callback_kwargs)
                    view.headers = view.default_response_headers
                    view.format_kwarg = view.get_format_suffix(
                        **callback_kwargs)
                    request.accepted_renderer, request.accepted_media_type = \
                        view.perform_content_negotiation(request)
                    version, scheme = view.determine_version(
                        request, *callback_args, **callback_kwargs)
                    request.version, request.versioning_scheme = version, scheme
                    response = view.options(
                        request, *callback_args, **callback_kwargs)
                    view.response = view.finalize_response(
                        request, response, *callback_args, **callback_kwargs)
                    return view.response

                return view.options(request, *callback_args, **callback_kwargs)

        return super(RulesMiddleware, self).process_view(
            request, callback, callback_args, callback_kwargs)

    def process_response(self, request, response):
        # Sets the CORS headers as appropriate.
        app = get_current_app(request)
        if app.cors_restricted:
            origin = request.META.get('HTTP_ORIGIN')
        else:
            origin = None
            response[ACCESS_CONTROL_ALLOW_HEADERS] = \
                ACCESS_CONTROL_ALLOW_HEADERS_ALLOWED
            response[ACCESS_CONTROL_ALLOW_ORIGIN] = "*"
            response[ACCESS_CONTROL_ALLOW_CREDENTIALS] = "true"

        if not origin:
            return super(RulesMiddleware, self).process_response(
                request, response)

        origin_parsed = six.moves.urllib.parse.urlparse(origin)
        if not origin_parsed.netloc:
            return super(RulesMiddleware, self).process_response(
                request, response)

        parts = request.get_host().split(':')
        host = parts[0].lower()
        port = parts[1] if len(parts) > 1 else None
        origin_parts = origin_parsed.netloc.split(':')
        origin_host = origin_parts[0].lower()
        origin_port = origin_parts[1] if len(origin_parts) > 1 else None
        if host != origin_host or port != origin_port:
            if origin_host.startswith('www.'):
                origin_host = origin_host[4:]
            if host == origin_host or host.endswith('.%s' % origin_host):
                patch_vary_headers(response, ['Origin'])
                response[ACCESS_CONTROL_ALLOW_HEADERS] = \
                    ACCESS_CONTROL_ALLOW_HEADERS_ALLOWED
                response[ACCESS_CONTROL_ALLOW_ORIGIN] = origin
                response[ACCESS_CONTROL_ALLOW_CREDENTIALS] = "true"
                # Patch cookies with `Domain=`
                self.patch_set_cookies(response, origin_host)
            else:
                logging.getLogger('django.security.SuspiciousOperation').info(
                    "request %s was not initiated by origin %s",
                    request.get_raw_uri(), origin)
        return super(RulesMiddleware, self).process_response(
            request, response)



# Copyright (c) 2020, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#pylint:disable=invalid-name,no-name-in-module,unused-import,import-error
#pylint:disable=unused-argument,bad-except-order,deprecated-method
#pylint:disable=import-outside-toplevel
from functools import WRAPPER_ASSIGNMENTS
import inspect, re

import six

from six.moves import http_cookies

try:
    from django.utils.decorators import available_attrs
except ImportError: # django < 3.0
    def available_attrs(fn):
        return WRAPPER_ASSIGNMENTS

try:
    from django.utils.encoding import python_2_unicode_compatible
except ImportError: # django < 3.0
    python_2_unicode_compatible = six.python_2_unicode_compatible


try:
    from django.apps import apps
    get_model = apps.get_model
except ImportError: # django < 1.8
    from django.db.models.loading import get_model


try:
    from django.utils.translation import gettext_lazy
except ImportError: # django < 3.0
    from django.utils.translation import ugettext_lazy as gettext_lazy


try:
    from django.urls import NoReverseMatch, reverse, reverse_lazy
except ImportError: # <= Django 1.10, Python<3.6
    from django.core.urlresolvers import NoReverseMatch, reverse, reverse_lazy
except ModuleNotFoundError: #pylint:disable=undefined-variable
    # <= Django 1.10, Python>=3.6
    from django.core.urlresolvers import NoReverseMatch, reverse, reverse_lazy

try:
    from django.urls import include, path, re_path
except ImportError: # <= Django 2.0, Python<3.6
    from django.conf.urls import include, url as re_path

    def path(route, view, kwargs=None, name=None):
        re_route = re.sub(
            r'<slug:([a-z\_]+)>', r'(?P<\1>[a-zA-Z0-9_\-\+\.]+)', route)
        return re_path(re_route, view, kwargs=kwargs, name=name)


def get_func_arg_names(func):
    try:
        return inspect.getfullargspec(func).args
    except AttributeError: # Python<3
        pass
    return inspect.getargspec(func)[0]


def get_model_class(full_name, settings_meta):
    """
    Returns a model class loaded from *full_name*. *settings_meta* is the name
    of the corresponding settings variable (used for error messages).
    """
    from django.core.exceptions import ImproperlyConfigured

    try:
        app_label, model_name = full_name.split('.')
    except ValueError:
        raise ImproperlyConfigured(
            "%s must be of the form 'app_label.model_name'" % settings_meta)
    model_class = get_model(app_label, model_name)
    if model_class is None:
        raise ImproperlyConfigured(
            "%s refers to model '%s' that has not been installed"
            % (settings_meta, full_name))
    return model_class


def is_authenticated(request):
    if callable(request.user.is_authenticated):
        return request.user.is_authenticated()
    return request.user.is_authenticated



# Copyright (c) 2024, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
CSV download view basics.
"""

from __future__ import unicode_literals

import csv
from io import BytesIO, StringIO

from django.http import HttpResponse
from django.views.generic import View
from rest_framework.request import Request

from ..compat import six
from ..mixins import AppMixin
from ..models import Rule
from ..utils import datetime_or_now
from ..api.rules import UserEngagementMixin


class CSVDownloadView(View):

    basename = 'download'
    filter_backends = []

    @property
    def headings(self):
        raise NotImplementedError

    @staticmethod
    def encode(text):
        if six.PY2:
            return text.encode('utf-8')
        return text

    @staticmethod
    def decorate_queryset(queryset):
        return queryset

    def filter_queryset(self, queryset):
        """
        Recreating a GenericAPIView.filter_queryset functionality here
        """
        # creating a DRF-compatible request object
        request = Request(self.request)
        for backend in list(self.filter_backends):
            queryset = backend().filter_queryset(request, queryset, self)
        return queryset

    def get(self, *args, **kwargs): #pylint: disable=unused-argument
        if six.PY2:
            content = BytesIO()
        else:
            content = StringIO()
        csv_writer = csv.writer(content)
        csv_writer.writerow([self.encode(head) for head in self.headings])
        qs = self.decorate_queryset(self.filter_queryset(self.get_queryset()))
        for record in qs:
            csv_writer.writerow(self.queryrow_to_columns(record))
        content.seek(0)
        resp = HttpResponse(content, content_type='text/csv')
        resp['Content-Disposition'] = \
            'attachment; filename="{}"'.format(
                self.get_filename())
        return resp

    def get_queryset(self):
        # Note: this should take the same arguments as for
        # Searchable and SortableListMixin in "extra_views"
        raise NotImplementedError

    def get_filename(self):
        return datetime_or_now().strftime(self.basename + '-%Y%m%d.csv')

    def queryrow_to_columns(self, record):
        raise NotImplementedError


class UserEngagementCSVView(UserEngagementMixin, AppMixin, CSVDownloadView):
    """
    Downloads the user engagement as a CSV file
    """

    @property
    def headings(self):
        if not hasattr(self, '_headings'):
            tags = set([])
            for rule in Rule.objects.get_rules(self.app):
                if rule.engaged:
                    tags |= set(rule.engaged.split(','))
            self._headings = ['user'] + list(tags)
        return self._headings

    def queryrow_to_columns(self, record):
        by_tags = {}
        for eng in record.engagements.all():
            tag = eng.slug.strip()
            if tag:
                by_tags[tag] = eng.last_visited

        row = [str(record)]
        for head in self.headings[1:]: # first column is user
            row += [by_tags.get(head, "")]
        return row



# Copyright (c) 2024, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime, json, logging, sys

from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Q
from django.utils.module_loading import import_string
from django.utils.timezone import utc
from pytz import timezone, UnknownTimeZoneError
from pytz.tzinfo import DstTzInfo

from .compat import six


LOGGER = logging.getLogger(__name__)


class JSONEncoder(json.JSONEncoder):

    def default(self, obj): #pylint: disable=method-hidden
        if hasattr(obj, 'isoformat'):
            return obj.isoformat()
        return super(JSONEncoder, self).default(obj)


def datetime_or_now(dtime_at=None):
    if not dtime_at:
        return datetime.datetime.utcnow().replace(tzinfo=utc)
    if dtime_at.tzinfo is None:
        dtime_at = dtime_at.replace(tzinfo=utc)
    return dtime_at


def update_context_urls(context, urls):
    if 'urls' in context:
        for key, val in six.iteritems(urls):
            if key in context['urls']:
                if isinstance(val, dict):
                    context['urls'][key].update(val)
                else:
                    # Because organization_create url is added in this mixin
                    # and in ``OrganizationRedirectView``.
                    context['urls'][key] = val
            else:
                context['urls'].update({key: val})
    else:
        context.update({'urls': urls})
    return context


def get_app_model():
    """
    Returns the Site model that is active in this project.
    """
    from . import settings
    try:
        return django_apps.get_model(settings.RULES_APP_MODEL)
    except ValueError:
        raise ImproperlyConfigured(
            "RULES_APP_MODEL must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured("RULES_APP_MODEL refers to model '%s'"\
" that has not been installed" % settings.RULES_APP_MODEL)


def get_app_serializer():
    """
    Returns the serializer model that is active in this project.
    """
    from . import settings
    try:
        return import_string(settings.APP_SERIALIZER)
    except ValueError:
        raise ImproperlyConfigured(
            "APP_SERIALIZER must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured("APP_SERIALIZER refers to model '%s'"\
" that has not been installed" % settings.APP_SERIALIZER)


def get_current_app(request=None):
    """
    Returns the default app for a site.
    """
    from . import settings
    if settings.DEFAULT_APP_CALLABLE:
        app = import_string(settings.DEFAULT_APP_CALLABLE)(request=request)
        LOGGER.debug("rules.get_current_app: '%s'", app)
    else:
        flt = Q(path_prefix__isnull=True)
        if request:
            request_path_parts = request.path.lstrip('/').split('/')
            if request_path_parts:
                flt = flt | Q(path_prefix='/%s' % request_path_parts[0])
        app = get_app_model().objects.filter(flt).order_by(
                'path_prefix', '-pk').first()
    return app


def get_current_enc_key(request=None):
    """
    Returns the default encoding key to encrypt forwarded sessions for a site.
    """
    from . import settings
    if settings.ENC_KEY_OVERRIDE:
        enc_key = getattr(sys.modules[__name__], '_ENC_KEY', None)
        if enc_key is None:
            try:
                enc_key = import_string(settings.ENC_KEY_OVERRIDE)
            except (ImportError, ModuleNotFoundError):
                enc_key = settings.ENC_KEY_OVERRIDE
            setattr(sys.modules[__name__], '_ENC_KEY', enc_key)
        if callable(enc_key):
            return enc_key(request=request)
        return enc_key
    return get_current_app(request=request).enc_key


def get_current_entry_point(request=None):
    """
    Returns the default forward entry point for a site.
    """
    from . import settings
    if settings.ENTRY_POINT_OVERRIDE:
        entry_point = getattr(sys.modules[__name__], '_ENTRY_POINT', None)
        if entry_point is None:
            try:
                entry_point = import_string(settings.ENTRY_POINT_OVERRIDE)
            except (ImportError, ModuleNotFoundError):
                entry_point = settings.ENTRY_POINT_OVERRIDE
            setattr(sys.modules[__name__], '_ENTRY_POINT', entry_point)
        if callable(entry_point):
            return entry_point(request=request)
        return entry_point
    return get_current_app(request=request).entry_point


def parse_tz(tzone):
    if issubclass(type(tzone), DstTzInfo):
        return tzone
    if tzone:
        try:
            return timezone(tzone)
        except UnknownTimeZoneError:
            pass
    return None



# Copyright (c) 2023, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime, logging

from dateutil.relativedelta import relativedelta
from django.contrib.auth import get_user_model
from django.db import transaction
from django.db.models import F, Q, Max, Count
from django.db.utils import IntegrityError
from django.utils.timezone import utc
from rest_framework.generics import (get_object_or_404, GenericAPIView,
    ListCreateAPIView, RetrieveUpdateDestroyAPIView, ListAPIView)
from rest_framework.response import Response
from rest_framework import serializers

from .serializers import (RuleSerializer, RuleRankUpdateSerializer,
    UserEngagementSerializer, EngagementsSerializer)
from ..docs import extend_schema, OpenApiResponse
from ..mixins import AppMixin
from ..models import Rule, Engagement
from ..utils import parse_tz, datetime_or_now


LOGGER = logging.getLogger(__name__)


class UpdateRuleSerializer(RuleSerializer):

    class Meta:
        model = Rule
        fields = ('rank', 'path', 'allow', 'is_forward', 'engaged')
        read_only_fields = ('path',)


class RuleMixin(AppMixin):

    model = Rule
    serializer_class = RuleSerializer
    lookup_field = 'path'
    lookup_url_kwarg = 'path'

    def get_queryset(self):
        return self.model.objects.get_rules(self.app)

    def insert_space(self, queryset, rank=1):
        queryset.filter(rank__gte=rank).update(rank=F('rank') + 1, moved=True)
        queryset.filter(moved=True).update(moved=False)


    def perform_create(self, serializer):
        # XXX adds rule at the top or bottom (False and True respectively).
        last_rank = 1
        if False:
            # We don't compact ranks after DELETE
            last_rank_qs = self.get_queryset().aggregate(Max('rank')).get(
                'rank__max', 0)
            # If the key exists and return None the default value is not applied
            if last_rank_qs is not None:
                last_rank = last_rank_qs + 1
        if not self.get_queryset().filter(
            path=serializer.validated_data['path']).exists():
            # If the path exists, we do not want to create a hole
            # when we are guarenteed to fail on insert.
            self.insert_space(self.get_queryset(), rank=last_rank)
        serializer.save(app=self.app, rank=last_rank)

    def perform_update(self, serializer):
        serializer.save(app=self.app)


class RuleListAPIView(RuleMixin, ListCreateAPIView):
    """
    Lists access rules

    Returns a list of {{PAGE_SIZE}} rules incoming HTTP requests
    are checked against.

    **Tags: rbac, broker, appmodel

    **Examples

    .. code-block:: http

        GET /api/proxy/rules HTTP/1.1

    responds

    .. code-block:: json

        {
            "count": 1,
            "next": null,
            "previous": null,
            "results": [
                {
                    "rank": 0,
                    "path": "/",
                    "allow": 1,
                    "is_forward": true,
                    "engaged": "app"
                }
            ]
        }
    """
    serializer_class = RuleSerializer

    def get_serializer_class(self):
        if self.request.method.lower() in ('patch',):
            return RuleRankUpdateSerializer
        return super(RuleListAPIView, self).get_serializer_class()

    def get_queryset(self):
        return super(RuleListAPIView, self).get_queryset().order_by('rank')

    @staticmethod
    def check_path(request):
        if 'path' not in request.data:
            request.data.update({'path': "/"})
        if not request.data['path'].startswith('/'):
            request.data['path'] = '/' + request.data['path']

    def perform_create(self, serializer):
        try:
            return super(RuleListAPIView, self).perform_create(serializer)
        except IntegrityError as err:
            if 'uniq' in str(err).lower():
                raise serializers.ValidationError({'detail':
                    "Rule with path prefix '%s' already exists."
                    % serializer.validated_data['path']})
            raise

    def post(self, request, *args, **kwargs):
        """
        Creates an access rule

        **Tags: rbac, broker, appmodel

        **Examples

        .. code-block:: http

            POST /api/proxy/rules HTTP/1.1

        .. code-block:: json

            {
                "rank": 0,
                "path": "/",
                "allow": 1,
                "is_forward": true,
                "engaged": ""
            }

        responds

        .. code-block:: json

            {
                "rank": 0,
                "path": "/",
                "allow": 1,
                "is_forward": true,
                "engaged": ""
            }
        """
        self.check_path(request)
        return self.create(request, *args, **kwargs)

    @extend_schema(responses={
      201: OpenApiResponse(RuleSerializer(many=True))})
    def patch(self, request, *args, **kwargs):
        """
        Updates order of rules

        When receiving a request like [{"newpos": 1, "oldpos": 3}],
        it will move the rule at position 3 to position 1, updating all
        rules ranks in-between.

        **Tags: rbac, broker, appmodel

        **Examples

        .. code-block:: http

            PATCH /api/proxy/rules HTTP/1.1

        .. code-block:: json

            {"updates": [
              {
                "newpos": 1,
                "oldpos": 3
               }
            ]}

        responds

        .. code-block:: json

            {
                "count": 1,
                "next": null,
                "previous": null,
                "results": [
                    {
                        "rank": 0,
                        "path": "/",
                        "allow": 1,
                        "is_forward": true,
                        "engaged": "app"
                    }
                ]
            }
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        with transaction.atomic():
            for move in serializer.validated_data['updates']:
                try:
                    oldrank = move['oldpos']
                    newrank = move['newpos']
                    queryset = self.get_queryset()
                    updated = queryset.get(rank=oldrank)
                    if newrank < oldrank:
                        queryset.filter(Q(rank__gte=newrank)
                                        & Q(rank__lt=oldrank)).update(
                            rank=F('rank') + 1, moved=True)
                    else:
                        queryset.filter(Q(rank__lte=newrank)
                                        & Q(rank__gt=oldrank)).update(
                            rank=F('rank') - 1, moved=True)
                    updated.rank = newrank
                    updated.moved = True
                    updated.save(update_fields=['rank', 'moved'])
                    queryset.filter(moved=True).update(moved=False)
                except Rule.DoesNotExist:
                    LOGGER.info("unable to move rule with rank=%d to rank=%d",
                        oldrank, newrank)

        return self.list(request, *args, **kwargs)

    def list(self, request, *args, **kwargs):
        # If we were to use `get_serializer`, it would return the calling
        # arguments serializer instead of the result serialize for PATCH calls.
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = RuleSerializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = RuleSerializer(queryset, many=True)
        return Response(serializer.data)


class RuleDetailAPIView(RuleMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieves an access rule

    **Tags: rbac, broker, appmodel

    **Examples

    .. code-block:: http

        GET /api/proxy/rules/app HTTP/1.1

    responds

    .. code-block:: json

        {
            "rank": 0,
            "path": "/app",
            "allow": 1,
            "is_forward": true,
            "engaged": ""
        }
    """
    serializer_class = UpdateRuleSerializer

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        path = self.kwargs.get(lookup_url_kwarg, "/")
        if not path.startswith('/'):
            path = '/' + path
        obj = get_object_or_404(queryset, path=path)
        return obj

    def put(self, request, *args, **kwargs):
        """
        Updates an access rule

        **Tags: rbac, broker, appmodel

        **Examples

        .. code-block:: http

            PUT /api/proxy/rules/app HTTP/1.1

        .. code-block:: json

            {
                "rank": 0,
                "allow": 1,
                "is_forward": true,
                "engaged": ""
            }

        responds

        .. code-block:: json

            {
                "rank": 0,
                "path": "/app",
                "allow": 1,
                "is_forward": true,
                "engaged": ""
            }
        """
        #pylint:disable=useless-super-delegation
        return super(RuleDetailAPIView, self).put(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        """
        Deletes an access rule

        **Tags: rbac, broker, appmodel

        **Examples

        .. code-block:: http

            DELETE /api/proxy/rules/app HTTP/1.1
        """
        #pylint:disable=useless-super-delegation
        return super(RuleDetailAPIView, self).delete(request, *args, **kwargs)

    def perform_destroy(self, instance):
        rank = instance.rank
        with transaction.atomic():
            instance.delete()
            # compacts ranks
            queryset = self.get_queryset()
            queryset.filter(rank__gte=rank).update(
                rank=F('rank') - 1, moved=True)
            queryset.filter(moved=True).update(moved=False)


class UserEngagementMixin(object):

    user_model = get_user_model()

    def get_queryset(self):
        return self.user_model.objects.order_by(
            '-last_login').prefetch_related('engagements')


class UserEngagementAPIView(UserEngagementMixin, ListAPIView):
    """
    Retrieves engagement metrics

    **Tags: rbac, broker, appmodel

    **Examples

    .. code-block:: http

        GET /api/proxy/engagement/users HTTP/1.1

    responds

    .. code-block:: json

        {
            "count": 2,
            "next": null,
            "previous": null,
            "results": [
              {
                "username": "alice",
                "engagements": ["app", "profile"]
              },
              {
                "username": "kenneth",
                "engagements": ["app", "billing"]
              }
            ]
        }
    """
    serializer_class = UserEngagementSerializer



class EngagementAPIView(AppMixin, GenericAPIView):

    serializer_class = EngagementsSerializer
    user_model = get_user_model()

    def get(self, request, *args, **kwargs):
        """
        Retrieves users engagement

        **Tags: rbac, broker, appmodel

        **Examples

        .. code-block:: http

            GET /api/proxy/engagement HTTP/1.1

        responds

        .. code-block:: json

            {
                "active_users": 10,
                "authentication": "enabled",
                "engagements": []
            }
        """
        #pylint:disable=unused-argument
        # https://docs.djangoproject.com/en/2.2/topics/db/aggregation/
        # #interaction-with-default-ordering-or-order-by
        engs = Engagement.objects.values('slug').annotate(
            count=Count('slug')).order_by('slug')
        nb_users = self.user_model.objects.count()
        engagement_stats = []
        for eng in engs:
            engagement_stats += [{
                'slug': eng['slug'],
                'count': eng['count'] * 100 / nb_users
            }]
        # Compute number active users for yesterday
        # XXX there should already be a function to compute that date range.
        yest_dt = datetime_or_now() - relativedelta(days=1)
        yest_start = datetime.datetime(year=yest_dt.year, month=yest_dt.month,
            day=yest_dt.day)
        yest_end = yest_start.replace(hour=23, minute=59, second=59,
            # https://stackoverflow.com/a/34593058/1491475
            microsecond=999999)
        timezone = self.request.GET.get('timezone')
        tz_ob = parse_tz(timezone)
        if not tz_ob:
            tz_ob = utc
        yest_start = tz_ob.localize(yest_start)
        yest_end = tz_ob.localize(yest_end)
        users = self.user_model.objects.filter(
            last_login__gt=yest_start, last_login__lt=yest_end).count()

        return Response(self.get_serializer({
            'authentication': self.app.authentication,
            'engagements': engagement_stats,
            'active_users': users}).data)



# Copyright (c) 2022, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
URLs for the resources API of rules.
"""

from ... import settings
from ...compat import path, re_path
from ...api.keys import (AppUpdateAPIView, GenerateKeyAPIView)
from ...api.rules import (RuleListAPIView, RuleDetailAPIView,
    UserEngagementAPIView, EngagementAPIView)
from ...api.sessions import GetSessionAPIView, GetSessionDetailAPIView

urlpatterns = [
    path('proxy/sessions/<slug:user>',
        GetSessionDetailAPIView.as_view(), name='rules_api_session_data'),
    path('proxy/sessions',
        GetSessionAPIView.as_view(), name='rules_api_session_data_base'),
    path('proxy/key',
        GenerateKeyAPIView.as_view(), name='rules_api_generate_key'),
    path('proxy/engagement/users',
        UserEngagementAPIView.as_view(), name='rules_api_user_engagement'),
    path('proxy/engagement',
        EngagementAPIView.as_view(), name='rules_api_engagement'),
    re_path(r'^proxy/rules/(?P<path>%s)$' % settings.PATH_RE,
        RuleDetailAPIView.as_view(), name='rules_api_rule_detail'),
    path('proxy/rules',
        RuleListAPIView.as_view(), name='rules_api_rule_list'),
    path('proxy',
        AppUpdateAPIView.as_view(), name='rules_api_app_detail'),
]



# Copyright (c) 2024, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Decorates URLs with permission checks.

This module implements `url` and `include` functions that apply a list
of permission checks before calling the underlying view.

The content of this file is inspired from
https://github.com/mila/django-urldecorators.git and patches
that were made in https://github.com:smirolo/django-urldecorators.git
for Django 2.0.
"""
from __future__ import unicode_literals

import logging
from functools import wraps

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.core.exceptions import ImproperlyConfigured

try:
    from django.urls.resolvers import URLPattern as DjangoRegexURLPattern
    from django.urls import URLResolver as DjangoRegexURLResolver
except ImportError: # Django<2.0
    from django.core.urlresolvers import (
        RegexURLPattern as DjangoRegexURLPattern,
        RegexURLResolver as DjangoRegexURLResolver
    )

from . import settings
from .compat import (available_attrs, get_func_arg_names,
    include as django_include, re_path as django_re_path, six)
from .perms import NoRuleMatch, check_matched, redirect_or_denied
from .utils import get_current_app


__all__ = ['include', 'url']


LOGGER = logging.getLogger(__name__)


class DecoratorMixin(object):
    """
    A mixin which adds support for decorating all resolved views
    """
    def __init__(self, *args, **kwargs):
        super(DecoratorMixin, self).__init__(*args, **kwargs)
        self.redirects = []
        self.decorators = []

    def resolve(self, path):
        match = super(DecoratorMixin, self).resolve(path)
        if not match:
            return match
        match.func = self.apply_decorators(match.func)
        return match

    def apply_decorators(self, func):
        """
        Adds checks and decorators into the call sequence.
        """
        for decorator_func in self.decorators:
            func = decorator_func(func)

        def decorator(view_func, redirects=[]):
            @wraps(view_func, assigned=available_attrs(view_func))
            def _wrapped_view(request, *view_args, **view_kwargs):
                LOGGER.debug("[DecoratorMixin._wrapped_view]"\
                    " request.path=%s, view_args=%s, view_kwargs=%s ...",
                    request.path, view_args, view_kwargs)
                try:
                    app = get_current_app(request)
                    #pylint:disable=unused-variable
                    redirect, matched, session = check_matched(request, app,
                        prefixes=settings.DEFAULT_PREFIXES)
                    if redirect:
                        return redirect_or_denied(request, redirect,
                            redirect_field_name=REDIRECT_FIELD_NAME)
                    return view_func(request, *view_args, **view_kwargs)
                except NoRuleMatch:
                    # No custom rules, so we will run the default set
                    # of rules in `redirects`.
                    pass
                for fail_func in redirects:
                    # filter out URL keyword arguments which the fail_func
                    # does not accept.
                    kwargs = {}
                    for name in get_func_arg_names(fail_func):
                        value = view_kwargs.get(name)
                        if value is not None:
                            kwargs.update({name: value})
                    LOGGER.debug(
                        "[decorator] calling %s(user=%s, kwargs=%s) ...",
                        fail_func.__name__, request.user, kwargs)
                    redirect = fail_func(request, **kwargs)
                    LOGGER.debug(
                    "[decorator] call returned %s(user=%s, kwargs=%s) => %s",
                        fail_func.__name__, request.user, kwargs, redirect)
                    if redirect:
                        return redirect_or_denied(request, redirect,
                            redirect_field_name=REDIRECT_FIELD_NAME)
                return view_func(request, *view_args, **view_kwargs)
            return _wrapped_view

        if func:
            return decorator(func, redirects=self.redirects)
        return decorator


class RegexURLPattern(DecoratorMixin, DjangoRegexURLPattern):
    """
    Django RegexURLPattern with support for decorating resolved views
    with permission checks.
    """


class RegexURLResolver(DecoratorMixin, DjangoRegexURLResolver):
    """
    Django RegexURLResolver with support for decorating resolved views
    with permission checks.
    """


def import_if_string(path):
    if not isinstance(path, six.string_types):
        return path
    try:
        dot = path.rindex('.')
    except ValueError:
        raise ImproperlyConfigured("%s isn't a valid module" % path)
    mod_name, obj_name = path[:dot], path[dot+1:]
    try:
        mod = __import__(mod_name, {}, {}, [''])
    except ImportError as err:
        raise ImproperlyConfigured('Error importing module %s: "%s"'
                                    % (mod_name, err))
    try:
        obj = getattr(mod, obj_name)
    except AttributeError:
        raise ImproperlyConfigured('Module "%s" does not define "%s"'
                                   % (mod_name, obj_name))
    return obj


def include(arg, namespace=None, app_name=None):
    try:
        #pylint:disable=unexpected-keyword-arg
        return django_include(arg, namespace=namespace, app_name=app_name)
    except TypeError:
        # Django2.0 removed the `app_name` parameter and asks to pass
        # it inside a tuple (`arg`, `app_name`).
        pass
    if not isinstance(arg, tuple):
        arg = (arg, app_name)
    return django_include(arg, namespace=namespace)


def url(regex, view, kwargs=None, name=None, prefix='',
        redirects=None, decorators=None):
    return re_path(regex, view, kwargs=kwargs, name=name, prefix=prefix,
        redirects=redirects, decorators=decorators)


def re_path(regex, view, kwargs=None, name=None, prefix='',
        redirects=None, decorators=None):
    """
    Extended `django.urls.re_path` with support to check checks before
    a view function is called.

    Example urls.py file:

        from rules.urldecorators import include, re_path

        urlpatterns = [
            re_path(r'^private/$', include('example.private.urls'),
                redirects=[rules.fail_authenticated]),
        ]
    """
    #pylint:disable=too-many-arguments
    if not redirects and not decorators:
        try:
            return django_re_path(regex, view, kwargs, name)
        except TypeError:  # Django<1.10
            #pylint:disable=too-many-function-args
            return django_re_path(regex, view, kwargs, name, prefix)
    rule = _url(regex, view, kwargs, name, prefix)
    rule.redirects = tuple([import_if_string(check)
        for check in redirects or []])
    rule.decorators = tuple([import_if_string(decorator)
        for decorator in reversed(decorators or [])])
    return rule

try:
    from django.urls.resolvers import RegexPattern

    def _url(regex, view, kwargs=None, name=None, prefix=''):
        #pylint:disable=unused-argument
        if not view:
            raise ImproperlyConfigured(
                "Empty URL pattern view name not permitted (for pattern %r)"
                % regex)
        if isinstance(view, (list, tuple)):
            # For include(...) processing.
            pattern = RegexPattern(regex, is_endpoint=False)
            urlconf_module, app_name, namespace = view
            return RegexURLResolver(
                pattern,
                urlconf_module,
                kwargs,
                app_name=app_name,
                namespace=namespace,
            )
        if callable(view):
            pattern = RegexPattern(regex, name=name, is_endpoint=True)
            return RegexURLPattern(pattern, view, kwargs, name)
        raise TypeError(
            "view must be a callable or a list/tuple in the case of include().")

except ImportError:
    def _url(regex, view, kwargs=None, name=None, prefix='',
             pattern=RegexURLPattern, resolver=RegexURLResolver):
        """
        Modified `django.urls.re_path` with allows to specify custom
        RegexURLPattern and RegexURLResolver classes.
        """
        #pylint:disable=too-many-arguments
        if isinstance(view, (list, tuple)):
            # For include(...) processing.
            return resolver(regex, view[0], kwargs, *view[1:])
        if isinstance(view, six.string_types):
            if not view:
                raise ImproperlyConfigured(
                "Empty URL pattern view name not permitted (for pattern %r)"
                % regex)
            if prefix:
                view = prefix + '.' + view
        return pattern(regex, view, kwargs, name)



# Copyright (c) 2024, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import unicode_literals

import json

from django.contrib.auth import get_user_model
from django.template.defaultfilters import slugify
from rest_framework import serializers

from .. import settings
from ..compat import gettext_lazy as _, six
from ..models import Rule
from ..utils import get_app_model


class EnumField(serializers.ChoiceField):
    """
    Treat a ``PositiveSmallIntegerField`` as an enum.
    """
    translated_choices = {}

    def __init__(self, choices, *args, **kwargs):
        self.translated_choices = {key: slugify(val) for key, val in choices}
        super(EnumField, self).__init__([(slugify(val), key)
            for key, val in choices],
            *args, **kwargs)

    def to_representation(self, value):
        if isinstance(value, list):
            result = [slugify(self.translated_choices.get(item, None))
                for item in value]
        else:
            result = slugify(self.translated_choices.get(value, None))
        return result

    def to_internal_value(self, data):
        if isinstance(data, list):
            result = [self.choices.get(item, None) for item in data]
        else:
            result = self.choices.get(data, None)
        if result is None:
            if not data:
                raise serializers.ValidationError(
                    _("This field cannot be blank."))
            raise serializers.ValidationError(
                _("'%(data)s' is not a valid choice."\
                " Expected one of %(choices)s.") % {
                    'data': data, 'choices': [str(choice)
                    for choice in six.iterkeys(self.choices)]})
        return result


class NoModelSerializer(serializers.Serializer):

    def create(self, validated_data):
        raise RuntimeError('`create()` should not be called.')

    def update(self, instance, validated_data):
        raise RuntimeError('`update()` should not be called.')


class AppSerializer(serializers.ModelSerializer):

    authentication = EnumField(
        choices=get_app_model().AUTH_TYPE, required=False,
        help_text=_("Restricted authentication and registration"))

    detail = serializers.CharField(required=False,
        help_text=_("Describes the result of the action"\
            " in human-readable form"))

    class Meta:
        model = get_app_model()
        fields = ('slug', 'entry_point', 'session_backend', 'authentication',
            'welcome_email', 'cors_restricted', 'detail')
        read_only_fields = ('slug', 'detail')

    @staticmethod
    def validate_entry_point(value):
        """
        Prevent unsafe URLs
        """
        parts = six.moves.urllib.parse.urlparse(value)
        if (not parts.netloc
            or parts.netloc.startswith('localhost')
            or parts.netloc.startswith('127.0.0.1')):
            raise serializers.ValidationError("Unsafe URLs are not allowed.")
        return value


class AppKeySerializer(serializers.ModelSerializer):
    """
    Used when a secret key is generated.
    """
    enc_key = serializers.CharField(read_only=True,
        help_text=_("Key used to decrypt the encoded session information."))

    class Meta:
        model = get_app_model()
        fields = ('enc_key',)


class RuleSerializer(serializers.ModelSerializer):

    allow = serializers.CharField(required=False, source='get_allow',
        help_text=_("Method applied to grant or deny access"))
    is_forward = serializers.BooleanField(required=False,
        help_text=_("When access is granted, should the request be forwarded"))
    engaged = serializers.CharField(required=False, allow_blank=True,
        help_text=_("Tags to check if it is the first time a user engages"))

    class Meta:
        model = Rule
        fields = ('rank', 'path', 'allow', 'is_forward', 'engaged')

    def validate(self, attrs):
        if 'get_allow' in attrs:
            parts = attrs.get('get_allow').split('/')
            try:
                rule_op = int(parts[0])
                if rule_op < 0 or rule_op >= len(settings.DB_RULE_OPERATORS):
                    raise ValueError("unknown rule %d" % rule_op)
                if len(parts) > 1:
                    params = json.loads(parts[1])
                    kwargs = {}
                    for key, dft in six.iteritems(
                            settings.RULE_OPERATORS[rule_op][2]):
                        if key in params:
                            kwargs[key] = params[key]
                        else:
                            kwargs[key] = dft
                    kwargs_encoded = json.dumps(kwargs)
                else:
                    kwargs_encoded = ""
                attrs['rule_op'] = rule_op
                attrs['kwargs'] = kwargs_encoded
            except ValueError as err:
                raise serializers.ValidationError(str(err))
            del attrs['get_allow']
        return super(RuleSerializer, self).validate(attrs)


class RuleRankSerializer(NoModelSerializer):

    oldpos = serializers.IntegerField(
        help_text=_("old rank for a rule in the list of rules"))
    newpos = serializers.IntegerField(
        help_text=_("new rank for the rule in the list of rules"))


class RuleRankUpdateSerializer(NoModelSerializer):

    updates = RuleRankSerializer(many=True)


class SessionDataSerializer(NoModelSerializer):

    forward_session = serializers.CharField(
        help_text=_("The session being forwarded"))
    forward_session_header = serializers.CharField(
        help_text=_("The HTTP header that encodes the session"))
    forward_url = serializers.CharField(
        help_text=_("The URL end point where the request is forwarded"))

    class Meta:
        fields = ('forward_session', 'forward_session_header', 'forward_url')
        read_only_fields = ('forward_session', 'forward_session_header',
            'forward_url')


class UsernameSerializer(NoModelSerializer):

    username = serializers.SerializerMethodField()

    class Meta:
        fields = ('username',)
        read_only_fields = ('username',)

    @staticmethod
    def get_username(request):
        return request.user.username


class UserEngagementSerializer(serializers.ModelSerializer):

    # Implementation Note:
    #   This serializer is used in a GET request but it will run
    #   a `rest_framework.UniqueValidator` if left unchecked when
    #   we run `_validate_examples` in the API doc generator.
    username = serializers.CharField(help_text=_("Username"))
    engagements = serializers.SerializerMethodField(
        help_text=_("Engagement tags for username"))

    class Meta:
        model = get_user_model()
        fields = ('username', 'engagements')

    @staticmethod
    def get_engagements(obj):
        engs = obj.engagements.all()
        user_tags = []
        for eng in engs:
            if eng.slug and not eng.slug.isspace():
                user_tags.append(eng.slug.strip())
        return list(set(user_tags))


class EngagementSerializer(NoModelSerializer):

    slug = serializers.CharField(
        help_text=_("Engagement tag"))
    count = serializers.IntegerField(
        help_text=_("Number of users that have engaged with this tag"))


class EngagementsSerializer(NoModelSerializer):

    engagements = EngagementSerializer(many=True,
        help_text=_("Engagement tags by user"))
    active_users = serializers.IntegerField(
        help_text=_("Number of users that have engaged with the app"))
    authentication = EnumField(
        choices=get_app_model().AUTH_TYPE, required=False,
        help_text=_("Restricted authentication and registration"))


class ValidationErrorSerializer(NoModelSerializer):
    """
    Details on why token is invalid.
    """
    detail = serializers.CharField(help_text=_("Describes the reason for"\
        " the error in plain text"))



# Copyright (c) 2021, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals

import logging

from django.conf import settings as django_settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.core.exceptions import PermissionDenied
from django.contrib.auth.views import redirect_to_login

from . import settings
from .compat import is_authenticated, six
from .models import Engagement, Rule
from .utils import datetime_or_now


LOGGER = logging.getLogger(__name__)


class NoRuleMatch(RuntimeError):

    def __init__(self, path):
        super(NoRuleMatch, self).__init__(
            "No access rules triggered by path '%s'" % path)


def _insert_url(request, redirect_field_name=REDIRECT_FIELD_NAME,
                inserted_url=django_settings.LOGIN_URL):
    '''Redirects to the *inserted_url* before going to the orginal
    request path.'''
    # This code is pretty much straightforward
    # from contrib.auth.user_passes_test
    path = request.build_absolute_uri()
    # If the login url is the same scheme and net location then just
    # use the path as the "next" url.
    login_scheme, login_netloc = six.moves.urllib.parse.urlparse(
        str(inserted_url))[:2]
    current_scheme, current_netloc = six.moves.urllib.parse.urlparse(path)[:2]
    if ((not login_scheme or login_scheme == current_scheme) and
        (not login_netloc or login_netloc == current_netloc)):
        path = request.get_full_path()
    # As long as *inserted_url* is not None, this call will redirect
    # anything (i.e. inserted_url), not just the login.
    return redirect_to_login(path, inserted_url, redirect_field_name)


def _get_accept_list(request):
    http_accept = request.META.get('HTTP_ACCEPT', '*/*')
    return [item.strip() for item in http_accept.split(',')]


def find_rule(request, app, prefixes=None):
    """
    Returns a tuple mad of the rule that was matched and a dictionnary
    of parameters that where extracted from the URL (i.e. :slug).
    """
    matched_rule = getattr(request, 'matched_rule', None)
    matched_params = getattr(request, 'matched_params', {})
    if matched_rule:
        # Use cached tuple.
        return (matched_rule, matched_params)
    params = {}
    request_path = request.path
    request_path_parts = [part for part in request_path.split('/') if part]
    for rule in Rule.objects.get_rules(app, prefixes=prefixes):
        params = rule.match(request_path_parts)
        if params is not None:
            LOGGER.debug(
                "matched %s with %s (rule=%d, forward=%s, params=%s)",
                request_path, rule.get_full_page_path(),
                rule.rule_op, rule.is_forward, params)
            return (rule, params)
        else:
            LOGGER.debug("match %s with %s ... no",
                '/'.join(request_path_parts), rule.get_full_page_path())
    return (None, {})


def redirect_or_denied(request, inserted_url,
                       redirect_field_name=REDIRECT_FIELD_NAME, descr=None):
    LOGGER.debug("redirect_or_denied((%s, %s), %s %s,"\
        " redirect_field_name=%s, descr=%s)", request.path, request.user,
        inserted_url, inserted_url.__class__, redirect_field_name, descr)
    http_accepts = _get_accept_list(request)
    if ('text/html' in http_accepts
        and isinstance(inserted_url, six.string_types)):
        return _insert_url(request,
            redirect_field_name=redirect_field_name, inserted_url=inserted_url)
    LOGGER.debug("Looks like an API call or no inserted url"\
        " (Accept: %s contains 'text/html' (%s),"\
        " inserted_url=%s is a string (%s)) => %s",
        request.META.get('HTTP_ACCEPT', '*/*'), 'text/html' in http_accepts,
        inserted_url, isinstance(inserted_url, six.string_types),
      ('PermissionDenied("%s")' % str(descr)) if descr else "PermissionDenied")
    if descr is None:
        descr = ""
    raise PermissionDenied(descr)


def engaged(rule, request=None):
    """
    Returns the last time the page was visited
    """
    last_visited = None
    if rule.engaged:
        last_tags = []
        datetime_stored = datetime_or_now()
        for tag in rule.engaged.split(','):
            obj, created = Engagement.objects.get_or_create(
                slug=tag, user=request.user,
                defaults={'last_visited': datetime_stored})
            if created:
                LOGGER.info(
                    "initial '%s' engagement%s", tag,
                    (" on %s" % request.path) if request is not None else "",
                    extra={'event': 'initial-engagement', 'request': request})
            elif last_visited:
                last_tags += [tag]
                last_visited = max(obj.last_visited, last_visited)
            else:
                last_tags += [tag]
                last_visited = obj.last_visited
        last_visited_tags = ','.join(last_tags)
        if last_tags and last_visited_tags != rule.engaged:
            last_visited = last_visited_tags
    return last_visited


def check_matched(request, app, prefixes=None, login_url=None):
    """
    Returns a tuple (response, forward, session) if the *request.path* can
    be matched otherwise raises a NoRuleMatch exception.
    """
    session = {}
    matched, params = find_rule(request, app, prefixes=prefixes)
    if matched:
        if matched.rule_op == Rule.ANY:
            if is_authenticated(request):
                last_visited = engaged(matched, request=request)
                session.update({'last_visited': last_visited})
            return (None, matched, session)

        redirect_url = None
        if not is_authenticated(request):
            LOGGER.debug("user is not authenticated")
            redirect_url = str(login_url or django_settings.LOGIN_URL)
        else:
            _, fail_func, defaults = settings.RULE_OPERATORS[matched.rule_op]
            kwargs = {}
            for key in defaults:
                if params and key in params:
                    kwargs.update({key: params[key]})
                else:
                    kwargs.update({key: defaults[key]})
            LOGGER.debug("[perms] calling %s(user=%s, kwargs=%s) ...",
                fail_func.__name__, request.user, kwargs)
            redirect_url = fail_func(request, **kwargs)
            LOGGER.debug("[perms] call returned %s(user=%s, kwargs=%s) => %s",
                fail_func.__name__, request.user, kwargs, redirect_url)
            if not redirect_url:
                redirect_url = None

        if not redirect_url:
            last_visited = engaged(matched, request=request)
            session.update({'last_visited': last_visited})
        return (redirect_url, matched if redirect_url is None else None,
            session)
    LOGGER.debug("unmatched %s", request.path)
    raise NoRuleMatch(request.path)


def check_permissions(request, app, login_url=None):
    """
    Returns a tuple (response, forward, session) if the *request.path* can
    be matched otherwise raises a PermissionDenied exception.
    """
    try:
        return check_matched(request, app, login_url=login_url)
    except NoRuleMatch as err:
        raise PermissionDenied(str(err))



# Copyright (c) 2024, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json, logging, re

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.sites.requests import RequestSite
from django.core.exceptions import FieldError, SuspiciousOperation
from django.db.models import Q
from django.http import HttpResponse, SimpleCookie
from django.template.response import TemplateResponse
from django.views.generic import UpdateView, TemplateView
import requests
from requests.exceptions import RequestException
from deployutils.apps.django.settings import SESSION_COOKIE_NAME

from .. import settings
from ..compat import get_model, http_cookies, reverse, six
from ..mixins import AppMixin, SessionDataMixin
from ..perms import (check_permissions as base_check_permissions,
    find_rule, redirect_or_denied)
from ..utils import (JSONEncoder, get_app_model, get_current_entry_point,
    update_context_urls)


LOGGER = logging.getLogger(__name__)


class DisallowedCookieName(SuspiciousOperation):
    """
    Cookie header contains invalid value
    """


class SessionProxyMixin(SessionDataMixin):
    """
    Proxy to the application

    Check permissions associated to the request and forwards request
    when appropriate.
    """
    redirect_field_name = REDIRECT_FIELD_NAME
    login_url = None

    def check_permissions(self, request):
        redirect_url, request.matched_rule, self.session = base_check_permissions(
            request, self.app, login_url=self.login_url)
        response = None
        if redirect_url:
            response = redirect_or_denied(
                request, redirect_url, self.redirect_field_name)
        if not response and self.session:
            # XXX Insert into self.request.session so we can use the same
            #     code on self-hosted templates.
            last_visited = self.session.get('last_visited', None)
            if last_visited:
                if not isinstance(last_visited, six.string_types):
                    last_visited = last_visited.isoformat()
                self.request.session.update({
                    'last_visited': last_visited})
        return (response, request.matched_rule.is_forward if request.matched_rule else False)

    def get_context_data(self, **kwargs):
        context = super(SessionProxyMixin, self).get_context_data(**kwargs)
        entry_point = get_current_entry_point(request=self.request)
        forward_url = '%s%s' % (entry_point, self.request.path)
        context.update({
            'forward_session': json.dumps(
                self.session, indent=2, cls=JSONEncoder),
            'forward_session_cookie': self.forward_session_header,
            'forward_url': forward_url,
        })
        return context

# Implementation Note:
# We cannot override ``dispatch()`` because djangorestframework happens
# to do user authentication there. Also doing so prevent injection
# of the edit tools.
# On the other hand, we cannnot blindly override `get`, `post`, etc. either
# because parents might not have implemented the method and it would result
# in 500 errors when incorrect HTTP requests are generated on the end points.
#   def dispatch(self, request, *args, **kwargs):
#       response = self.conditional_forward(request)
#       if response:
#           return response
#       # If we get here, the request was forwarded and we got an exception
#       # or the request must be handled locally.
#       return super(SessionProxyMixin, self).dispatch(request, *args, **kwargs)

    def options(self, request, *args, **kwargs):
        # With CORS the browser strips the Authentication header yet
        # it expects a 200 OK response.
        self.session = {}
        request.matched_rule, request.matched_params = find_rule(
            request, self.app)
        if request.matched_rule and request.matched_rule.is_forward:
            try:
                return self.fetch_remote_page()
            except RequestException as err:
                return self.forward_error(err)
        return self.forward_to_parent(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        response = self.conditional_forward(request)
        if response:
            return response
        return self.forward_to_parent(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        response = self.conditional_forward(request)
        if response:
            return response
        return self.forward_to_parent(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        response = self.conditional_forward(request)
        if response:
            return response
        return self.forward_to_parent(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        response = self.conditional_forward(request)
        if response:
            return response
        return self.forward_to_parent(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        response = self.conditional_forward(request)
        if response:
            return response
        return self.forward_to_parent(request, *args, **kwargs)

    def forward_to_parent(self, request, *args, **kwargs):
        if request.method.lower() in self.http_method_names:
            handler = getattr(super(SessionProxyMixin, self),
                request.method.lower(), self.http_method_not_allowed)
        else:
            handler = self.http_method_not_allowed
        return handler(request, *args, **kwargs)

    def forward_error(self, err):
        context = self.get_context_data()
        context.update({'err': str(err)})
        return TemplateResponse(
            request=self.request,
            template='rules/forward_error.html',
            context=context,
            content_type='text/html',
            status=503)

    def conditional_forward(self, request):
        response, forward = self.check_permissions(request)
        if response:
            return response
        if forward:
            try:
                return self.fetch_remote_page()
            except RequestException as err:
                return self.forward_error(err)
        return None

    def fetch_remote_page(self):
        """
        Forward the request to the remote site after adjusting session
        information and request headers.

        Respond with the remote site response after adjusting session
        information and response headers.
        """
        entry_point = get_current_entry_point(request=self.request)
        forward_url = '%s%s' % (entry_point, self.request.path) # XXX
        requests_args = self.translate_request_args(self.request)
        if LOGGER.getEffectiveLevel() == logging.DEBUG:
            LOGGER.debug("\"%s %s (Fwd to %s)\" with session %s,"\
                " updated headers: %s",
                self.request.method, self.request.path, entry_point,
                self.session, requests_args)
        else:
            LOGGER.info("\"%s %s (Fwd to %s)\"", self.request.method,
                self.request.path, entry_point, extra={
                    'event': 'http_forward', 'fwd_to': entry_point,
                    'request': self.request})
        response = requests.request(
            self.request.method, forward_url, timeout=settings.TIMEOUT,
            **requests_args)
        return self.translate_response(response)

    def translate_request_args(self, request):
        #pylint:disable=too-many-statements
        requests_args = {'allow_redirects': False, 'headers': {}}
        cookies = SimpleCookie()
        try:
            for key, value in six.iteritems(request.COOKIES):
                LOGGER.debug("attempt to set %s=%s", key, value)
                if key.lower() == 'samesite':
                    continue
                cookies[key] = value
            if (self.app.session_backend and
                self.app.session_backend != self.app.JWT_SESSION_BACKEND):
                cookies[SESSION_COOKIE_NAME] = self.session_cookie_string
            else:
                # Cookies, as opposed to Authorization header have multiple
                # purposes, so we keep the SESSION_COOKIE_NAME (if exists)
                # regardless of the backend used to transport the proxy session.
                pass
        except http_cookies.CookieError as err:
            # Some is messing up with the 'Cookie' header. This sometimes
            # happen with bots trying to set 'Max-Age' or other reserved words.
            raise DisallowedCookieName(str(err))

        #pylint: disable=maybe-no-member
        # Something changed in `SimpleCookie.output` that creates an invalid
        # cookie string starting with a spacel or there are more strident
        # checks in the requests module (2.11.1) that prevents passing
        # a cookie string starting with a space.
        cookie_string = cookies.output(header='', sep=';').strip()

        # Retrieve the HTTP headers from a WSGI environment dictionary.  See
        # https://docs.djangoapp.com/en/dev/ref/request-response/\
        #    #django.http.HttpRequest.META
        headers = {}
        for key, value in six.iteritems(request.META):
            key_upper = key.upper()
            if key_upper.startswith('HTTP_'):
                key_upper = key_upper[5:].replace('_', '-')
                if key_upper == 'AUTHORIZATION':
                    # We donot want to inadvertently forward the credentials
                    # to connect to the proxy when the session backend is
                    # using cookies.
                    pass
                if key_upper == 'COOKIE':
                    headers[key_upper] = cookie_string
                else:
                    # Some servers don't like when you send the requesting host
                    # through but we do anyway. That's how the nginx reverse
                    # proxy is configured as well.
                    headers[key_upper] = value
            elif key_upper in ('CONTENT_TYPE', 'CONTENT_LENGTH'):
                headers[key.replace('_', '-')] = value

        site = RequestSite(request)
        if 'HOST' not in headers:
            headers.update({'HOST': site.domain})
        if 'X-FORWARDED-FOR' not in headers:
            headers.update({'X-FORWARDED-FOR': site.domain})
        if 'X-REAL-IP' not in headers:
            headers.update({'X-REAL-IP': request.META.get('REMOTE_ADDR', None)})
        if 'COOKIE' not in headers:
            headers.update({'COOKIE': cookie_string})

        if self.app.session_backend and \
            self.app.session_backend == self.app.JWT_SESSION_BACKEND:
            jwt_token = self.session_jwt_string
            headers.update({'AUTHORIZATION': 'Bearer %s' % jwt_token})

        if request.META.get(
                'CONTENT_TYPE', '').startswith('multipart/form-data'):
            if request.FILES:
                requests_args['files'] = request.FILES
            data = {}
            for key, val in six.iteritems(request.POST):
                data.update({key:val})
            requests_args['data'] = data
        else:
            requests_args['data'] = request.body

        params = request.GET.copy()

        # If there's a content-length header from Django, it's probably
        # in all-caps and requests might not notice it, so just remove it.
        for key in list(headers.keys()):
            if key.lower() == 'content-length':
                del headers[key]
            elif key.lower() == 'content-type' and request.META.get(
                'CONTENT_TYPE', '').startswith('multipart/form-data'):
                del headers[key]

        requests_args['headers'] = headers
        requests_args['params'] = params
        return requests_args

    def translate_response(self, response):
        proxy_response = HttpResponse(
            response.content, status=response.status_code)
        if 'set-cookie' in response.headers:
            # Here we have to decode the Set-Cookie ourselves because
            # requests will pack the set-cookie headers under the same key,
            # comma separated, which comma SimpleCookie.load() will append
            # to the path in the Morsel class (ie. Path=/,).
            # This of course results in the browser not sending the cookie
            # back to us later on.
            #pylint: disable=protected-access
            if six.PY2:
                set_cookie_lines \
                    = response.raw._original_response.msg.getallmatchingheaders(
                        'Set-Cookie')
            else:
                # We implement our own search here because
                # ``getallmatchingheaders`` is broken in Python3
                # (see https://bugs.python.org/issue5053)
                set_cookie_lines = []
                for line, data in six.iteritems(
                        response.raw._original_response.msg):
                    if line.lower() == 'set-cookie':
                        set_cookie_lines.append(line + ': ' + data)

            set_cookies_cont = ''
            set_cookies = []
            for line in set_cookie_lines:
                # The parsing is complicated by the fact that
                # ``getallmatchingheaders`` will return continuation lines
                # as an entry in the list.
                if line.startswith('Set-Cookie'):
                    if set_cookies_cont:
                        set_cookies += [set_cookies_cont]
                    set_cookies_cont = line[11:].strip()
                else:
                    set_cookies_cont += line.strip()
            if set_cookies_cont:
                set_cookies += [set_cookies_cont]

            intercepted_cookies = SimpleCookie()
            for cookie_text in set_cookies:
                intercepted_cookies.load(cookie_text)
            excluded_cookies = set([
                'sessionid', # XXX hardcoded
            ])
            for key in excluded_cookies:
                if key in intercepted_cookies:
                    del intercepted_cookies[key]
            #pylint: disable=maybe-no-member
            proxy_response.cookies.update(intercepted_cookies)

        excluded_headers = set([
            'set-cookie', # Previously parsed.
            # Hop-by-hop headers
            # ------------------
            # Certain response headers should NOT be just tunneled through.
            # For more info, see:
            # http://www.w3.org/Protocols/rfc2616/rfc2616-sec13.html#sec13.5.1
            'connection', 'keep-alive', 'proxy-authenticate',
            'proxy-authorization', 'te', 'trailers', 'transfer-encoding',
            'upgrade',

            # Although content-encoding is not listed among the hop-by-hop
            # headers, it can cause trouble as well.  Just let the server set
            # the value as it should be.
            'content-encoding',

            # Since the remote server may or may not have sent the content
            # in the same encoding as Django will, let Django worry about
            # what the length should be.
            'content-length',
        ])
        for key, value in six.iteritems(response.headers):
            if key.lower() in excluded_headers:
                continue
            proxy_response[key] = value
        return proxy_response


class SessionProxyView(SessionProxyMixin, AppMixin, TemplateView):

    pass


class AppDashboardView(AppMixin, UpdateView):
    """
    Update a ``App``'s fields associated to the proxy dashboard
    (i.e. entry point and encryption key).
    """

    model = get_app_model()
    fields = ('entry_point',)
    template_name = 'rules/app_dashboard.html'

    def get_object(self, queryset=None):
        return self.app

    def get_context_data(self, **kwargs):
        #pylint:disable=too-many-locals
        context = super(AppDashboardView, self).get_context_data(**kwargs)
        rules = []
        for idx, rule in enumerate(settings.RULE_OPERATORS):
            short_name, _, defaults = rule
            parms = defaults.copy()
            look = re.search(r'%\((\S+)\)s', short_name)
            if look:
                cls_path = look.group(1)
                cls = get_model(cls_path)
                cls_name = cls_path.split('.')[-1].lower()
                try:
                    queryset = cls.objects.filter(
                        organization=self.app.account, is_active=True)
                except FieldError:
                    # Support optional null ``organization`` foreign keys
                    # for ``saas.RoleDescription`` global roles.
                    try:
                        queryset = cls.objects.filter(
                            Q(organization__isnull=True) |
                            Q(organization=self.app.account))
                    except FieldError:
                        queryset = cls.objects.all()
                for obj in queryset:
                    parms.update({cls_name: obj.slug})
                    identifier = '%d/%s' % (idx, json.dumps(parms))
                    parms.update()
                    descr = short_name % {cls_path: obj.title}
                    rules += [(identifier, descr)]
            else:
                rules += [(str(idx), short_name)]
        context.update({
            'site_available_at_url': self.request.build_absolute_uri('/'),
            'organization': self.app.account,
            'rules': rules
        })
        return context


class UserEngagementView(AppMixin, TemplateView):

    template_name = 'rules/engagement.html'

    def get_context_data(self, **kwargs):
        context = super(UserEngagementView, self).get_context_data(**kwargs)
        update_context_urls(context, {
            'rules': {
               'user_engagement_download': reverse(
                   'rules_user_engagement_download')
        }})
        return context



# Copyright (c) 2023, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from random import choice

from rest_framework.response import Response
from rest_framework.generics import GenericAPIView, RetrieveUpdateAPIView

from ..compat import gettext_lazy as _
from ..docs import extend_schema
from ..mixins import AppMixin
from ..utils import get_app_model, get_app_serializer
from .serializers import AppKeySerializer


class GenerateKeyAPIView(AppMixin, GenericAPIView):
    """
    Generate a new key to communicate between proxy and backend service.
    """
    http_method_names = ['post']
    model = get_app_model()
    serializer_class = AppKeySerializer

    @extend_schema(request=None)
    def post(self, request, *args, **kwargs):
        """
        Rotates session encoding key

        Rotates the key used to encode the session information forwarded
        to the application entry point.

        **Tags: rbac, broker, appmodel

        **Examples

        .. code-block:: http

            POST /api/proxy/key HTTP/1.1

        responds

        .. code-block:: json

            {
              "enc_key": "********"
            }
        """
        #pylint:disable=unused-argument
        self.app.enc_key = "".join([
                choice("abcdefghijklmnopqrstuvwxyz0123456789!@#$%^*-_=+")
                for idx in range(16)]) #pylint: disable=unused-variable
        self.app.save()
        return Response(self.get_serializer().to_representation(self.app))


class AppUpdateAPIView(AppMixin, RetrieveUpdateAPIView):
    """
    Retrieves forward end-point

    Returns the URL endpoint to which requests passing the access rules
    are forwarded to, and the format in which the session information
    is encoded.

    When running tests, you can retrieve the actual session information
    for a specific user through the
    :ref:`session API endpoint <#retrieveGetSessionDetail>`.

    **Tags: rbac, broker, appmodel

    **Examples

    .. code-block:: http

        GET /api/proxy HTTP/1.1

    responds

    .. code-block:: json

        {
          "slug": "cowork",
          "entry_point": "https://cowork.herokuapp.com/",
          "session_backend": 1
        }
    """
    model = get_app_model()
    serializer_class = get_app_serializer()

    def get_object(self):
        return self.app

    def put(self, request, *args, **kwargs):
        """
        Updates forward end-point

        Updates the URL endpoint to which requests passing the access rules
        are forwarded to and/or the format in which the session information
        is encoded.

        **Tags: rbac, broker, appmodel

        **Examples

        .. code-block:: http

            PUT /api/proxy HTTP/1.1

        .. code-block:: json

            {
              "entry_point": "https://cowork.herokuapp.com/",
              "session_backend": 1
            }

        responds

        .. code-block:: json

            {
              "slug": "cowork",
              "entry_point": "https://cowork.herokuapp.com/",
              "session_backend": 1
            }
        """
        #pylint:disable=useless-super-delegation
        return super(AppUpdateAPIView, self).put(request, *args, **kwargs)

    def perform_update(self, serializer):
        serializer.save()
        serializer.instance.detail = _("Update successful.")
